#include <vortex-python/common.hpp>
#include <vortex-python/memory.hpp>
#include <vortex-python/exception.hpp>

#include <vortex/acquire.hpp>

template<typename C>
static void setup_acquire_config(py::class_<C>& c) {
    c.def(py::init());

    RO_ACC(shape);

    RW_ACC(samples_per_record);
    RW_ACC(records_per_block);
}

static void bind_null(py::module& m) {
    {
        using C = vortex::null_acquisition_t<uint16_t>;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "NullAcquisition");

        c.def(py::init());
        
        FXN_GIL(initialize);

        FXN_GIL(prepare);
        FXN_GIL(start);
        FXN_GIL(stop);

        c.def("next", [](C& o, const vortex::cpu_view_t<C::output_element_t>& buffer, size_t id) {
            py::gil_scoped_release();
            return o.next(id, buffer);
        }, "buffer"_a, "id"_a = 0);

        RO_ACC(config);
    }

    {
        using C = vortex::null_acquisition_t<uint16_t>::config_t;
        auto c = py::class_<C>(m, "NullAcquisitionConfig");

        setup_acquire_config(c);

        RW_ACC(channels_per_sample);
    }
}

static void bind_file(py::module& m) {
    {
        using C = vortex::file_acquisition_t<uint16_t>;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "FileAcquisition");

        c.def(py::init());
        c.def(py::init<std::shared_ptr<spdlog::logger>>(), "logger"_a);

        FXN_GIL(initialize);

        FXN_GIL(prepare);
        FXN_GIL(start);
        FXN_GIL(stop);

        c.def("next", [](C& o, const vortex::cpu_view_t<C::output_element_t>& buffer, size_t id) {
            py::gil_scoped_release();
            return o.next(id, buffer);
        }, "buffer"_a, "id"_a = 0);

        RO_ACC(config);
    }

    {
        using C = vortex::file_acquisition_t<uint16_t>::config_t;
        auto c = py::class_<C>(m, "FileAcquisitionConfig");

        setup_acquire_config(c);

        RW_ACC(channels_per_sample);

        RW_VAR(path);
        RW_VAR(loop);
    }
}

#if defined(VORTEX_ENABLE_ALAZAR)

PYBIND11_MAKE_OPAQUE(std::vector<vortex::alazar_acquisition_t::config_t::option_t>);
PYBIND11_MAKE_OPAQUE(std::vector<vortex::acquire::input_t>);

namespace pybind11 {
    namespace detail {
        template <typename... Ts>
        struct type_caster<vortex::acquire::clock_t<Ts...>> : variant_caster<vortex::acquire::clock_t<Ts...>> {};
        template <typename... Ts>
        struct type_caster<vortex::acquire::trigger_t<Ts...>> : variant_caster<vortex::acquire::trigger_t<Ts...>> {};
        template <typename... Ts>
        struct type_caster<vortex::acquire::option_t<Ts...>> : variant_caster<vortex::acquire::option_t<Ts...>> {};
    }
}

template<typename C>
static void setup_alazar_config(py::class_<C>& c) {
    RW_VAR(device);
    RW_VAR(clock);
    RW_VAR(trigger);
    RW_VAR(inputs);
    RW_VAR(options);

    RW_VAR(resampling);

    RW_VAR(acquire_timeout);
    RW_VAR(stop_on_error);

    RO_ACC(channels_per_sample);

    RO_ACC(bytes_per_multisample);
    RO_ACC(channel_mask);

    RW_ACC(samples_per_second);
    RO_ACC(samples_per_second_is_known);

    RO_ACC(recommended_minimum_records_per_block);

    FXN(validate);
}

template<typename view_t, typename C>
static void setup_alazar_acquisition(py::class_<C, std::shared_ptr<C>>& c) {
    c.def(py::init());
    c.def(py::init<std::shared_ptr<spdlog::logger>>(), "logger"_a);

    RO_ACC(config);

    FXN_GIL(initialize);

    FXN_GIL(prepare);
    FXN_GIL(start);
    FXN_GIL(stop);
    RO_ACC(running);

    c.def("next", [](C& o, const view_t& buffer) {

        py::gil_scoped_release gil;
        return o.next(buffer);

    }, "buffer"_a);
    c.def("next", [](C& o, size_t id, const view_t& buffer) {

        py::gil_scoped_release gil;
        return o.next(id, buffer);

    }, "id"_a, "buffer"_a);

    c.def("next_async", [](C& o, const view_t& buffer, typename C::callback_t callback) {

        py::gil_scoped_release gil;
        o.next_async(buffer, std::move(callback));

    }, "buffer"_a, "callback"_a);
    c.def("next_async", [](C& o, size_t id, const view_t& buffer, typename C::callback_t callback) {

        py::gil_scoped_release gil;
        o.next_async(id, buffer, std::move(callback));

    }, "id"_a, "buffer"_a, "callback"_a);

}

template<typename C>
static std::string alazar_inner_repr(const C& v){
    std::string result = fmt::format("(\n\
    shape = {},\n\
    channels_per_sample = {},\n\
    samples_per_record = {},\n\
    records_per_block = {},\n\
    device = {},\n\
    clock = {},\n\
    trigger = {},\n\
    inputs = {},\n\
    options = {},\n\
    acquire_timeout = {},\n\
    stop_on_error = {},\n\
    bytes_per_multisample = {},\n\
    channel_mask = {},\n\
    samples_per_second = {},\n\
    samples_per_second_is_known = {},\n\
    recommended_minimum_records_per_block = {},",
             PY_REPR(v.shape()),
             v.channels_per_sample(),
             v.samples_per_record(),
             v.records_per_block(),
             PY_REPR(v.device),
             PY_REPR(v.clock),
             PY_REPR(v.trigger),
             PY_REPR(v.inputs),
             PY_REPR(v.options),
             PY_REPR(v.acquire_timeout),
             PY_REPR(v.stop_on_error),
             v.bytes_per_multisample(),
             PY_REPR(v.channel_mask()),
             v.samples_per_second(),
             PY_REPR(v.samples_per_second_is_known()),
             v.recommended_minimum_records_per_block()
             );
    return result;
}
static void bind_alazar(py::module& m) {

    py::bind_vector<std::vector<vortex::alazar_acquisition_t::config_t::option_t>>(m, "VectorOption")
        .def("__repr__", [](const std::vector<vortex::alazar_acquisition_t::config_t::option_t>& v){
                             return vector_repr(v, "VectorOption");
                         });
    py::bind_vector<std::vector<vortex::acquire::input_t>>(m, "VectorInput")
        .def("__repr__", [](const std::vector<vortex::acquire::input_t>& v){
                             return vector_repr(v, "VectorInput");
                         });

    {
        using C = vortex::acquire::clock::internal_t;
        auto c = py::class_<C>(m, "InternalClock");

        c.def(py::init());
        c.def(py::init([](size_t samples_per_second) -> C {
            return { samples_per_second };
               }), "samples_per_second"_a);
        c.def("__repr__", [](const C& v){ return fmt::format("InternalClock(samples_per_second={})", PY_REPR(v.samples_per_second));});

        RW_VAR(samples_per_second);

        SHALLOW_COPY();
    }
    {
        using C = vortex::acquire::clock::external_t;
        auto c = py::class_<C>(m, "ExternalClock");

        c.def(py::init());
        c.def("__repr__", [](const C& v){
                  return fmt::format("ExternalClock(coupling={}, dual={}, edge={}, level_ratio={:0.2f})",
                         PY_REPR(v.coupling), PY_REPR(v.dual), PY_REPR(v.edge), v.level_ratio);});

        RW_VAR(level_ratio);
        RW_VAR(coupling);
        RW_VAR(edge);
        RW_VAR(dual);

        SHALLOW_COPY();
    }
    {
        using C = vortex::acquire::trigger::single_external_t;
        auto c = py::class_<C>(m, "SingleExternalTrigger");

        c.def(py::init());
        c.def(py::init([](size_t range_millivolts=2500,
                          float level_ratio= 0.09f,
                          size_t delay_samples=80,
                          vortex::alazar::trigger_slope_t slope=vortex::alazar::trigger_slope_t::positive,
                          vortex::alazar::coupling_t coupling=vortex::alazar::coupling_t::DC)
                          -> C {return {range_millivolts, level_ratio, delay_samples, slope, coupling};}),
              "range_millivolts"_a = 2500,
              "level_ratio"_a = 0.09f,
              "delay_samples"_a = 80,
              "slope"_a = vortex::alazar::trigger_slope_t::positive,
              "coupling"_a = vortex::alazar::coupling_t::DC);
        c.def("__repr__",
              [](const C& v){
                  return fmt::format("SingleExternalTrigger(range_millivolts={}, level_ratio={:0.2f}, delay_samples={}, slope={}, coupling={})",
                                     PY_REPR(v.range_millivolts), v.level_ratio, PY_REPR(v.delay_samples), PY_REPR(v.slope), PY_REPR(v.coupling));
              });
        
        RW_VAR(range_millivolts);
        RW_VAR(level_ratio);
        RW_VAR(delay_samples);
        RW_VAR(slope);
        RW_VAR(coupling);
    }
    {
        using C = vortex::acquire::trigger::dual_external_t;
        auto c = py::class_<C>(m, "DualExternalTrigger");

        c.def(py::init());
        c.def(py::init([](size_t range_millivolts=2500,
                          std::array<float, 2> level_ratios= {0.09f, 0.09f},
                          size_t delay_samples=80,
                          vortex::alazar::trigger_slope_t initial_slope=vortex::alazar::trigger_slope_t::positive,
                          vortex::alazar::coupling_t coupling=vortex::alazar::coupling_t::DC)
                          -> C {return {range_millivolts, level_ratios, delay_samples, initial_slope, coupling};}),
              "range_millivolts"_a = 2500,
              "level_ratios"_a = std::array<float, 2> ({0.09f, 0.09f}),
              "delay_samples"_a = 80,
              "initial_slope"_a = vortex::alazar::trigger_slope_t::positive,
              "coupling"_a = vortex::alazar::coupling_t::DC);
        c.def("__repr__",
              [](const C& v){
                  return fmt::format("DualExternalTrigger(range_millivolts={}, level_ratios={}, delay_samples={}, initial_slope={}, coupling={})",
                                     PY_REPR(v.range_millivolts), PY_REPR(v.level_ratios), PY_REPR(v.delay_samples), PY_REPR(v.initial_slope), PY_REPR(v.coupling));
              });
        
        RW_VAR(range_millivolts);
        RW_VAR(level_ratios);
        RW_VAR(delay_samples);
        RW_VAR(initial_slope);
        RW_VAR(coupling);
    }

    {
        using C = vortex::acquire::input_t;
        auto c = py::class_<C>(m, "Input");

        c.def(py::init());
        c.def(py::init([](vortex::alazar::channel_t channel) -> C {
            return { channel };
        }));
        c.def("__repr__", [](const C& v){
                              return fmt::format("Input(channel={}, range_millivolts={}, impedance_ohms={}, coupling={}, bytes_per_samples={})",
                                 PY_REPR(v.channel), PY_REPR(v.range_millivolts), PY_REPR(v.impedance_ohms), PY_REPR(v.coupling), PY_REPR(v.bytes_per_sample()));
                          });

        RW_VAR(channel);
        RW_VAR(range_millivolts);
        RW_VAR(impedance_ohms);
        RW_VAR(coupling);

        RO_ACC(bytes_per_sample);

        SHALLOW_COPY();
    }

    {
        using C = vortex::acquire::option::auxio_trigger_out_t;
        auto c = py::class_<C>(m, "AuxIOTriggerOut");

        c.def(py::init());
        c.def(py::self == py::self);
        c.def("__repr__", [](const C& v){return "AuxIOTriggerOut";});

        SHALLOW_COPY();
    }
    {
        using C = vortex::acquire::option::auxio_clock_out_t;
        auto c = py::class_<C>(m, "AuxIOClockOut");

        c.def(py::init());
        c.def(py::self == py::self);
        c.def("__repr__", [](const C& v){return "AuxIOClockOut";});

        SHALLOW_COPY();
    }

    {
        using C = vortex::acquire::option::auxio_pacer_out_t;
        auto c = py::class_<C>(m, "AuxIOPacerOut");

        c.def(py::init());
        c.def(py::self == py::self);
        c.def("__repr__", [](const C& v){return fmt::format("AuxIOPacerOut(divider={})", v.divider);});

        RW_VAR(divider);

        SHALLOW_COPY();
    }
    {
        using C = vortex::acquire::option::oct_ignore_bad_clock_t;
        auto c = py::class_<C>(m, "OCTIgnoreBadClock");

        c.def(py::init());
        c.def(py::self == py::self);
        c.def("__repr__", [](const C& v){
                              return fmt::format("OCTIgnoreBadClock(good_seconds={}, bad_seconds={})",
                                                 v.good_seconds, v.bad_seconds);
                          });
        RW_VAR(good_seconds);
        RW_VAR(bad_seconds);

        SHALLOW_COPY();
    }

    {
        using T = vortex::alazar_acquisition_t;
        {
            using C = T::config_t::device_t;
            auto c = py::class_<C>(m, "AlazarConfigDevice");

            c.def(py::init());
            c.def(py::init<U32, U32>());
            c.def("__repr__", [](const C& v){
                                  return fmt::format("AlazarConfigDevice(system_index={}, board_index={})",
                                                     PY_REPR(v.system_index), PY_REPR(v.board_index));
                              });

            RW_VAR(system_index);
            RW_VAR(board_index);

            SHALLOW_COPY();
        }

        {
            using C = T::config_t;
            auto c = py::class_<C>(m, "AlazarConfig");
            c.def("__repr__", [](const C& v){return fmt::format("AlazarConfig{}\n)", alazar_inner_repr(v));});

            setup_acquire_config(c);
            setup_alazar_config(c);

            SHALLOW_COPY();
        }

        {
            using C = T;
            auto c = py::class_<C, std::shared_ptr<C>>(m, "AlazarAcquisition");

            setup_alazar_acquisition<vortex::cpu_view_t<typename T::output_element_t>>(c);
        }
    }
 
    {
        using T = vortex::alazar_fft_acquisition_t;

        {
            using C = T::config_t;
            auto c = py::class_<C>(m, "AlazarFFTConfig");
            c.def("__repr__", [](const C& v) {return fmt::format("AlazarFFTConfig{}\n)", alazar_inner_repr(v)); });

            setup_acquire_config(c);
            setup_alazar_config(c);

            RW_VAR(fft_length);
            RW_VAR_XT(spectral_filter);
            RW_VAR_XT(background);
            RW_VAR(include_time_domain);

            RO_ACC(samples_per_ascan);
            RW_ACC(ascans_per_block);

            SHALLOW_COPY();
        }

        {
            using C = T;
            auto c = py::class_<C, std::shared_ptr<C>>(m, "AlazarFFTAcquisition");

            setup_alazar_acquisition<vortex::cpu_view_t<typename T::output_element_t>>(c);
        }
    }

#if defined(VORTEX_ENABLE_ALAZAR_GPU)
    {
        using T = vortex::alazar_gpu_acquisition_t;
        {
            using C = T::config_t;
            auto c = py::class_<C>(m, "AlazarGPUConfig");
            c.def("__repr__", [](const C& v){
                                  return fmt::format("AlazarGPUConfig{}\n    gpu_device = {},\n)",
                                                     alazar_inner_repr(v), v.gpu_device_index);
                              });

            setup_acquire_config(c);
            setup_alazar_config(c);

            RW_VAR(gpu_device_index);

            SHALLOW_COPY();
        }

        {
            using C = T;
            auto c = py::class_<C, std::shared_ptr<C>>(m, "AlazarGPUAcquisition");

            setup_alazar_acquisition<vortex::alazar::alazar_view_t<typename T::output_element_t>>(c);
        }
    }
#endif

}

#endif

void bind_acquire(py::module& root) {

    auto m = root.def_submodule("acquire");

    bind_null(m);
    bind_file(m);

#if defined(VORTEX_ENABLE_ALAZAR)
    bind_alazar(m);
#endif

}
