#include <vortex/process.hpp>

#include <vortex-python/common.hpp>
#include <vortex-python/memory.hpp>

template<typename C>
static void setup_processor_config(py::class_<C>& c) {
    RW_ACC(records_per_block);
    RW_ACC(samples_per_record);
    
    RW_ACC(ascans_per_block);
    RO_ACC(samples_per_ascan);

    RW_VAR(channel);

    RO_ACC(input_shape);    
    RO_ACC(output_shape);

    RW_VAR(average_window);

    RW_VAR_XT(spectral_filter);
    RW_VAR_XT(resampling_samples);
    
    RW_VAR(enable_ifft);
    RW_VAR(enable_log10);
}

static void bind_null(py::module& m) {
    using T = vortex::null_processor_t;
    {
        using C = T::config_t;
        auto c = py::class_<C>(m, "NullProcessorConfig");

        c.def(py::init());
        setup_processor_config(c);

        SHALLOW_COPY();
    }

    {
        using C = T;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "NullProcessor");

        c.def(py::init());

        RO_ACC(config);

        FXN_GIL(initialize);
    }
}

static void bind_copy(py::module& m) {
    using T = vortex::copy_processor_t<uint16_t, int8_t>;
    {
        using C = T::config_t;
        auto c = py::class_<C>(m, "CopyProcessorConfig");

        c.def(py::init());
        
        RW_ACC(records_per_block);
        RW_ACC(samples_per_record);

        RW_ACC(ascans_per_block);
        RO_ACC(samples_per_ascan);

        RW_VAR(channel);

        RW_VAR(slots);

        RO_ACC(input_shape);
        RO_ACC(output_shape);

        RW_VAR(sample_slice);
        RW_VAR(sample_transform);

        SHALLOW_COPY();
    }

    {
        using C = T;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "CopyProcessor");

        c.def(py::init());
        c.def(py::init<std::shared_ptr<spdlog::logger>>());

        RO_ACC(config);

        FXN_GIL(initialize);
        FXN_GIL(change);

        c.def("next", [](C& o,
            const vortex::cpu_view_t<const typename T::input_element_t>& input_buffer,
            const vortex::cpu_view_t<typename T::output_element_t>& output_buffer) {

                py::gil_scoped_release gil;
                o.next(input_buffer, output_buffer);

        }, "input_buffer"_a, "output_buffer"_a);
    }
}

#if defined(VORTEX_ENABLE_CUDA)

static void bind_cuda(py::module& m) {
    using T = vortex::cuda_processor_t<uint16_t, int8_t>;
    {
        using C = T::config_t;
        auto c = py::class_<C>(m, "CUDAProcessorConfig");

        c.def(py::init());
        setup_processor_config(c);

        RW_VAR(device);
        RW_VAR(slots);

        FXN(validate);

        SHALLOW_COPY();
    }

    {
        using C = T;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "CUDAProcessor");

        c.def(py::init());
        c.def(py::init<std::shared_ptr<spdlog::logger>>());

        RO_ACC(config);

        FXN_GIL(initialize);
        FXN_GIL(change);

        c.def("next", [](C& o,
            const vortex::cuda::cuda_view_t<const typename T::input_element_t>& input_buffer,
            const vortex::cuda::cuda_view_t<typename T::output_element_t>& output_buffer,
            bool append_history) {

            py::gil_scoped_release gil;
            o.next(input_buffer, output_buffer, append_history);

        }, "input_buffer"_a, "output_buffer"_a, "append_history"_a = true);

    }
}

#endif

void bind_process(py::module& root) {
    auto m = root.def_submodule("process");

    bind_null(m);
    bind_copy(m);

#if defined(VORTEX_ENABLE_CUDA)
    bind_cuda(m);
#endif

}
