#pragma once

#include <numeric>
#include <complex>

#include <cuda_runtime_api.h>
#include <cuComplex.h>

#include <fmt/format.h>

// NOTE: no namespace shorthand here so NVCC can include this header
namespace vortex {
    namespace cuda {

        class exception : public std::runtime_error {
        public:
            using runtime_error::runtime_error;
        };

        namespace detail {

            template<typename T>
            struct cuda_complex_t {};

            template<>
            struct cuda_complex_t<float> { using type = cuFloatComplex; };
            template<>
            struct cuda_complex_t<double> { using type = cuDoubleComplex; };


            template<typename T>
            struct cuda_type_t { using type = T; };
            template<>
            struct cuda_type_t<std::complex<float>> { using type = cuda_complex_t<float>::type; };
            template<>
            struct cuda_type_t<std::complex<double>> { using type = cuda_complex_t<double>::type; };

            template<typename... Args>
            void handle_error(cudaError_t error, const std::string& msg, Args... args) {
                if (!error) {
                    return;
                }

                // emit an exception
#if FMT_VERSION >= 80000
                auto user_msg = fmt::format(fmt::runtime(msg), args...);
#else
                auto user_msg = fmt::format(msg, args...);
#endif
                auto error_msg = fmt::format("{}: ({}) {}", user_msg, error, cudaGetErrorName(error));
                throw exception(error_msg);
            }
        }

        template<typename T>
        using complex = typename detail::cuda_complex_t<T>::type;

        template<typename T>
        using device_type = typename detail::cuda_type_t<T>::type;

    }
}